/* Parameter for handling audio buffers in scripts\
0508 and later
*/



BufParameter : Parameter {
	var buffer;
	makeGui { | gui, adapterEnvir |
		var label, bufsink, numbox, menu, adapter;
		label = SCStaticText(gui, Rect(0,0,100,20)).string_(name);
		bufsink = SCDragSink(gui, Rect(0,0,140,20))
			.string_(" NO BUFFER ")
			.background_(Color(0.3, 0.6, 0.9))
			.font_(Font("Helvetica", 10))
			.canReceiveDragHandler_({ this.canReceiveDragHandler })
			.receiveDragHandler_({ this.receiveDragHandler });
		numbox = SCNumberBox(gui, Rect(0,0,50,20));
		numbox.value = script.envir[name];
		menu = SCPopUpMenu(gui, Rect(0, 0, 100, 20));
		menu.items = Samples.default.buffers.keys.asArray.sort;
		menu.font = Font("Helvetica", 10);
		menu.action = {
			var bnum;
			bnum = Samples[menu.items[menu.value]] ? 0;
			action.(Samples[menu.items[menu.value]] ? 0);
			script.changed(name, bnum);
		};
		menu.keyDownAction =  { | me, char, mod, unicode |
			// do not react to space key bubbled from top view!
			//	if (char == $ , { me.valueAction = me.value + 1; ^me });
			switch (char,
				$\r, { me.valueAction = me.value + 1; },
				$\n, { me.valueAction = me.value + 1; },
				3.asAscii, { me.valueAction = me.value + 1; },
				{
					switch (unicode,
						16rF700, { me.valueAction = me.value + 1; },
						16rF703, { me.valueAction = me.value + 1; },
						16rF701, { me.valueAction = me.value - 1; },
						16rF702, { me.valueAction = me.value - 1; }
					)
				}
			)
		};
		adapter = { | sender, whatChanged, args |
			switch (whatChanged,
				\bufList, {
					{ menu.items = args; }.defer;
					// must set selection to actual buffer on menu! : ...
				},
				\deleted, {
					[this, \deletedBuffer, args].postln; // since deleted: 
					script.set(name, 0); // set to empty buffer
				}
			);
		};
		Samples.default.addDependant(adapter);
		menu.onClose = { Samples.default.removeDependant(adapter); };
		// update slider knob color when midi responder activation state changes: 
		// TODO: examine if BufParameter is midiable, use label color to indicate active status
		// 
/*		adapterEnvir[midiName] = { | isOn |
			if (isOn) {
//				menu.knobColor = Color.red;
			}{
//				menu.knobColor = Color.clear;
			}
		};
*/		adapterEnvir[name] = { | bufnum |
			{
				var bufname;
				numbox.value = bufnum = bufnum.round(1).asInteger;
				buffer = Samples.default.buffers.detect {|b| b.bufnum == bufnum };
//				[this, name, bufnum, buffer].postln;
				if (buffer.notNil) {
					bufname = (buffer.path ? " --- ").fileSymbol;
					bufsink.string = bufname ++ "(" ++
					((buffer.numFrames ? 0) / (buffer.sampleRate ? 44100)).round(0.01).asString
					++ ")";
					menu.value = menu.items.indexOf(bufname) ? 0;
				};
			}.defer;
		};
		// update label color when midi responder activation state changes: 
		adapterEnvir[midiName] = { | isOn |
			if (isOn) {
				label.background = Color(1, 0.5, 0.5, 0.5);
			}{
				label.background = Color.clear;
			}
		};
		// perform update of your gui after constructoin, thereby setting gui items
		script.changed(name, script.envir[name]);
	}
	canReceiveDragHandler { | dragsink, numbox, gui |
		var object;
		object = SCView.currentDrag;
		if (object.isKindOf(MIDIResponder)) { ^true };
		^object.isKindOf(Buffer);
	}
	receiveDragHandler {
		if (SCView.currentDrag.isKindOf(MIDIResponder)) {
			this.setMIDIResponder(SCView.currentDrag);
		}{
			action.(SCView.currentDrag.bufnum ? 0);
		}
	}
	// --------------- MIDI:
	// attempt at midiability: 
/*	canAutoMidiSetup {
		^false;
	}*/
	// attempt at midiability ...
	activateMidi {
		super.activateMidi;
		this.update(Samples.default, \bufList);
		Samples.default.addDependant(this);
	}
	deactivateMidi {
		super.deactivateMidi;
		Samples.default.removeDependant(this);
	}
	update { | who, what |
		// buf parameters add themselves to Samples so that they can update 
		// their menu and specs when a buffer is loaded or freed. 
		// this here makes it possible to scroll through the list of Samples.default.buffers
		// when midi is received on this parameter 
//		[this, who, what].postln;
//		if (what == \samples) {
		spec.maxval = (who.buffers.size - 1) max: 0;
//		}
	}
}
