/* IZ 050713

*/

LinkedBus : Bus {
	var <readers; /* dict (event) of processes will be mapped / set to read from this bus
		Association structure:
		value: the process, key: the name of the parameter to set / map in that process
		This bus will be allocated as soon as a running reader is added, or just before any
		linke reader is started */
	var <activeReaders; 	// array of currently running processes that read from this bus
						// bus will be freed when there are no more active readers
	var <writers; /* dict (event) of processes that are configured to write to this bus
		their outputs are set to this bus whenever it is allocated */

	*new { arg rate=\audio, numChannels=1, server;
		^this.newCopyArgs(rate, nil, numChannels, server ?? { Server.default }, (),
			IdentitySet.new, ()).init;
	}

	// free yourself when server quits to avoid errors on CmdPeriod
	init {
		server.onQuit({ index = nil }, server);
	}

	// add a process to read from this bus at control named param
	addReader { | reader, param = \in |
		readers[reader] = param;
		if (index.notNil) {
			if (rate == \audio) { reader.set(param, index) } { reader.map(param, index) }
 		}
	}

	removeReader { | reader |
		readers.removeAt(reader);
		this.stop(reader);
	}

	addWriter { | writer, param = \out |
		writers[writer] = param;
		if (index.notNil) { writer.set(param, index) };
	}

	removeWriter { | writer | writers.removeAt(writer); }

	start { | reader |
//		[\starting, this, reader, writers].postln;
		if (index.isNil) {
			this.alloc;
			writers.keysValuesDo { | w, param |
//				[w, \setting, param, index].postln;
				w.set(param, index); };
			if (rate == \audio) {
				readers.keysValuesDo { | r, param | r.set(param, index); }
			}{
				readers.keysValuesDo { | r, param | r.map(param, index); }
			}
		};
		activeReaders.add(reader);
	}

	stop { | reader |
		activeReaders.remove(reader);
		if (activeReaders.size == 0 and: { index.notNil }) {
			// inlining 'free' method code, without removal of numChannel value
			if(rate == \audio,{
				server.audioBusAllocator.free(index);
			},{
				server.controlBusAllocator.free(index);
			});
			index = nil;
		}
	}
}
